"""Shared test utilities."""

import datetime as dt
from typing import List

from magodo import MagicTodoMixin
from magodo.spells import DEFAULT_TODO_SPELLS, add_o_prefix, remove_o_prefix
from magodo.types import AbstractTodo, LineSpell, TodoSpell


CREATE_DATE = dt.datetime.strptime("1900-01-01", "%Y-%m-%d").date()
DONE_DATE = dt.datetime.strptime("2022-02-07", "%Y-%m-%d").date()
METADATA = {"ctime": "HHMM", "dtime": "HHMM"}

MOCK_TODO_KWARGS = {
    "metadata": METADATA,
    "create_date": CREATE_DATE,
    "done_date": DONE_DATE,
}


class MagicTodo(MagicTodoMixin):
    """The default MagicTodo class."""

    to_line_spells: List[LineSpell] = [add_o_prefix]
    todo_spells: List[TodoSpell] = DEFAULT_TODO_SPELLS
    from_line_spells: List[LineSpell] = [remove_o_prefix]


def assert_todos_equal(actual: AbstractTodo, expected: AbstractTodo) -> None:
    """Helper function for asserting that one Todo is equal to another."""
    assert (
        actual.desc == expected.desc
    ), f"{actual.desc!r} != {expected.desc!r}"
    assert (
        actual.priority == expected.priority
    ), f"{actual.priority!r} != {expected.priority!r}"
    assert (
        actual.create_date == expected.create_date
    ), f"{actual.create_date!r} != {expected.create_date!r}"
    assert (
        actual.done_date == expected.done_date
    ), f"{actual.done_date!r} != {expected.done_date!r}"
    assert (
        actual.projects == expected.projects
    ), f"{actual.projects!r} != {expected.projects!r}"
    assert (
        actual.contexts == expected.contexts
    ), f"{actual.contexts!r} != {expected.contexts!r}"
    assert (
        actual.done == expected.done
    ), f"{actual.done!r} != {expected.done!r}"
    assert (
        actual.metadata == expected.metadata
    ), f"{actual.metadata!r} != {expected.metadata!r}"



"""Contains the MagicTodoMixin class definition."""

from __future__ import annotations

import abc
import datetime as dt
from functools import total_ordering
from typing import Any, List, Tuple, Type, TypeVar
import uuid

from eris import ErisError, Err, Ok, Result

from ._todo import Todo, TodoMixin
from .types import LineSpell, Metadata, Priority, TodoSpell, ValidateSpell


M = TypeVar("M", bound="MagicTodoMixin")


@total_ordering
class MagicTodoMixin(TodoMixin, abc.ABC):
    """Mixin class that implements the Todo protocol."""

    to_line_spells: List[LineSpell] = []
    todo_spells: List[TodoSpell] = []
    from_line_spells: List[LineSpell] = []
    validate_spells: List[ValidateSpell] = []

    def __init__(self: M, todo: Todo):
        self._todo = todo
        self.todo = self.cast_todo_spells(todo)

    @property
    def ident(self) -> str:
        """Unique identifier."""
        return str(uuid.uuid4())

    @classmethod
    def from_line(cls: Type[M], line: str) -> Result[M, ErisError]:
        """Converts a string into a MagicTodo object."""
        err: Err[Any, ErisError]
        if error := cls.cast_validate_spells(line).err():
            err = Err(f"Failed spell validation for this todo: line={line!r}")
            return err.chain(error)

        line = cls.cast_from_line_spells(line)
        todo_result = Todo.from_line(line)

        if isinstance(todo_result, Err):
            err = Err(
                "Failed to construct basic Todo object inside of MagicTodo."
            )
            return err.chain(todo_result)

        todo = todo_result.ok()

        return Ok(cls(todo))

    def to_line(self: M) -> str:
        """Converts this MagicTodo back to a string."""
        line = self.todo.to_line()
        line = self.cast_to_line_spells(line)
        return line

    @classmethod
    def cast_todo_spells(cls: Type[M], todo: Todo) -> Todo:
        """Casts all spells associated with this MagicTodo on `todo`."""
        new_todo = todo.new()
        for todo_spell in cls.todo_spells:
            new_todo = todo_spell(new_todo)

        return new_todo

    @classmethod
    def cast_from_line_spells(cls: Type[M], line: str) -> str:
        """Casts all from_line spells on `line`."""
        for line_spell in cls.from_line_spells:
            line = line_spell(line)
        return line

    def cast_to_line_spells(self: M, line: str) -> str:
        """Casts all to_line spells on `line`."""
        for line_spell in self.to_line_spells:
            line = line_spell(line)
        return line

    @classmethod
    def cast_validate_spells(
        cls: Type[M], line: str
    ) -> Result[None, ErisError]:
        """Casts all spells associated with this MagicTodo on `todo`."""
        for validate_spell in cls.validate_spells:
            new_todo_result = validate_spell(line)
            if isinstance(new_todo_result, Err):
                err: Err[Any, ErisError] = Err(
                    f"The {validate_spell.__name__!r} validation spell failed"
                    " while attempting to validate this line."
                )
                return err.chain(new_todo_result)

        return Ok(None)

    def new(self: M, **kwargs: Any) -> M:
        """Creates a new Todo using the current Todo's attrs as defaults."""
        return type(self)(self.todo.new(**kwargs))

    @property
    def contexts(self: M) -> Tuple[str, ...]:  # noqa: D102
        return self.todo.contexts

    @property
    def create_date(self: M) -> dt.date | None:  # noqa: D102
        return self.todo.create_date

    @property
    def desc(self: M) -> str:  # noqa: D102
        return self.todo.desc

    @property
    def done_date(self: M) -> dt.date | None:  # noqa: D102
        return self.todo.done_date

    @property
    def done(self: M) -> bool:  # noqa: D102
        return self.todo.done

    @property
    def metadata(self: M) -> Metadata:  # noqa: D102
        return self.todo.metadata

    @property
    def priority(self: M) -> Priority:  # noqa: D102
        return self.todo.priority

    @property
    def projects(self: M) -> Tuple[str, ...]:  # noqa: D102
        return self.todo.projects



"""Test that we can create custom MagicTodo classes."""

from __future__ import annotations

from typing import List, Type

from pytest import mark

from magodo import MagicTodoMixin, Todo
from magodo.types import AbstractMagicTodo, TodoSpell

from .shared import MOCK_TODO_KWARGS, MagicTodo, assert_todos_equal


params = mark.parametrize


def to_test_line(line: str) -> str:
    """Prepends 'test |' to `line`."""
    return "test | " + line


def from_test_line(line: str) -> str:
    """Removes 'test | ' from `line`."""
    return line.replace("test | ", "")


class LineTodo(MagicTodoMixin):
    """MagicTodo that adds 'test | ' to the beginning of each line."""

    from_line_spells = [from_test_line]
    to_line_spells = [to_test_line]
    todo_spells: List[TodoSpell] = []

    ident: str = "LINE TODO"


@params(
    "line,magic_todo_type,pre_inner_todo,expected_inner_todo",
    [
        (
            "x:1030 2022-01-12 Some done todo.",
            MagicTodo,
            Todo(
                desc="x:1030 Some done todo.",
                metadata={"x": "1030"},
                done=True,
            ),
            Todo(
                desc="Some done todo. | ctime:HHMM dtime:HHMM",
                done=True,
            ),
        ),
        (
            "x:1030 2022-01-12 Some done todo with a @ctx.",
            MagicTodo,
            Todo(
                desc="x:1030 Some done todo with a @ctx.",
                contexts=("ctx",),
                metadata={"x": "1030"},
                done=True,
            ),
            Todo(
                desc="Some done todo with a ctx. | @ctx ctime:HHMM dtime:HHMM",
                contexts=("ctx",),
                done=True,
            ),
        ),
    ],
)
def test_magic_todo(
    line: str,
    magic_todo_type: Type[AbstractMagicTodo[Todo]],
    pre_inner_todo: Todo,
    expected_inner_todo: Todo,
) -> None:
    """Test the MagicTodo.from_line() function."""
    from_line_mtodo = magic_todo_type.from_line(line).unwrap()
    from_line_mtodo = from_line_mtodo.new(**MOCK_TODO_KWARGS)

    pre_inner_todo = pre_inner_todo.new(**MOCK_TODO_KWARGS)
    from_todo_mtodo = magic_todo_type(pre_inner_todo).new(**MOCK_TODO_KWARGS)
    post_inner_todo = from_todo_mtodo.todo.new(**MOCK_TODO_KWARGS)
    expected_inner_todo = expected_inner_todo.new(**MOCK_TODO_KWARGS)

    assert_todos_equal(post_inner_todo, expected_inner_todo)

    assert repr(from_line_mtodo) == repr(from_todo_mtodo)


@params("line", ["foo bar baz", "test | foo bar baz"])
def test_line_todo(line: str) -> None:
    """Test that to_line and from_line spells work."""
    todo = LineTodo.from_line(line).unwrap()
    assert todo.desc == "foo bar baz"
    assert todo.to_line() == "test | foo bar baz"



"""Shared utility functions / constants / classes."""

from __future__ import annotations

import datetime as dt
from typing import Final

from .types import Priority


CONTEXT_PREFIX: Final = "@"
DATE_FMT: Final = "%Y-%m-%d"
DEFAULT_PRIORITY: Final[Priority] = "O"
PROJECT_PREFIX: Final = "+"
PUNCTUATION: Final = ",.?!;"
RE_DATE: Final = r"[1-9][0-9][0-9][0-9]-[0-1][0-9]-[0-3][0-9]"
TODO_PREFIXES: Final = ("x ", "x:", "o ")


def to_date(yyyymmdd: str) -> dt.date:
    """Helper function for constructing a date object."""
    return dt.datetime.strptime(yyyymmdd, DATE_FMT).date()


def from_date(date: dt.date) -> str:
    """Helper function for converting a date object to a string."""
    return date.strftime(DATE_FMT)


def is_metadata_tag(word: str) -> bool:
    """Predicate that tells us if `word` is a metadata tag or not."""
    kv = word.split(":", maxsplit=1)
    return bool(len(kv) == 2 and kv[1] and not kv[1].startswith(":"))


def is_prefix_tag(prefix: str, word: str) -> bool:
    """Predicate that tells us if a word is prefixed by `prefix`."""
    return (
        word.startswith(prefix)
        and not word.startswith(prefix + prefix)
        and len(prefix) < len(word)
    )


def is_context_tag(word: str) -> bool:
    """Returns True if `word` is a context tag."""
    return is_prefix_tag(CONTEXT_PREFIX, word)


def is_project_tag(word: str) -> bool:
    """Returns True if `word` is a project tag."""
    return is_prefix_tag(PROJECT_PREFIX, word)


def is_any_prefix_tag(word: str) -> bool:
    """Returns True if `word` is either a context or project tag."""
    return is_context_tag(word) or is_project_tag(word)


def is_any_tag(word: str) -> bool:
    """Returns True if `word` is a project, context, or metadata."""
    return is_any_prefix_tag(word) or is_metadata_tag(word)



"""The catch-all module for the tests.data package."""

from pathlib import Path
from typing import List, cast

from pytest import param


def get_all_todo_paths() -> List[Path]:
    """Returns the full path of every todo.txt file in the todos directory."""
    todo_dir_path = Path(__file__).absolute().parent / "todos"

    result = []

    for txt_or_dir in todo_dir_path.glob("*"):
        if txt_or_dir.is_file() and txt_or_dir.suffix != ".txt":
            continue

        P = cast(Path, param(txt_or_dir, id=txt_or_dir.stem))
        result.append(P)

    return result



"""Contains all of magodo's default MagicTodo spell functions."""

from __future__ import annotations

import datetime as dt
from functools import lru_cache as cache
from typing import Callable, Final, Iterable, List, MutableSequence, cast

from eris import ErisResult, Err, Ok
from typist import literal_to_list

from ._shared import (
    CONTEXT_PREFIX,
    PROJECT_PREFIX,
    PUNCTUATION,
    TODO_PREFIXES,
    is_any_prefix_tag,
    is_any_tag,
    is_metadata_tag,
)
from .types import LineSpell, Priority, T, TodoSpell, ValidateSpell


def register_todo_spell_factory(
    mut_spell_registry: MutableSequence[TodoSpell],
) -> Callable[[TodoSpell], TodoSpell]:
    """Factory for decorators used to register spell functions."""

    def register_spell(spell: TodoSpell) -> TodoSpell:
        mut_spell_registry.append(spell)
        return spell

    return register_spell


def register_line_spell_factory(
    mut_spell_registry: MutableSequence[LineSpell],
) -> Callable[[LineSpell], LineSpell]:
    """Factory for decorators used to register spell functions."""

    def register_spell(spell: LineSpell) -> LineSpell:
        mut_spell_registry.append(spell)
        return spell

    return register_spell


def register_validate_spell_factory(
    mut_spell_registry: MutableSequence[ValidateSpell],
) -> Callable[[ValidateSpell], ValidateSpell]:
    """Factory for decorators used to register spell functions."""

    def register_spell(spell: ValidateSpell) -> ValidateSpell:
        mut_spell_registry.append(spell)
        return spell

    return register_spell


DEFAULT_TODO_SPELLS: List[TodoSpell] = []
todo_spell = register_todo_spell_factory(DEFAULT_TODO_SPELLS)

DEFAULT_TO_LINE_SPELLS: List[LineSpell] = []
to_line_spell = register_line_spell_factory(DEFAULT_TO_LINE_SPELLS)

DEFAULT_FROM_LINE_SPELLS: List[LineSpell] = []
from_line_spell = register_line_spell_factory(DEFAULT_FROM_LINE_SPELLS)

DEFAULT_VALIDATE_SPELLS: List[ValidateSpell] = []
validate_spell = register_validate_spell_factory(DEFAULT_VALIDATE_SPELLS)

O_PREFIX: Final = "o "


@cache
def todo_prefixes() -> tuple[str, ...]:
    """Returns all valid todo prefixes."""
    result = list(TODO_PREFIXES)
    for P in cast(List[str], literal_to_list(Priority)):
        result.append(f"({P}) ")
    return tuple(result)


@validate_spell
def validate_prefix(line: str) -> ErisResult[None]:
    """Handles / validatees a magic todo's prefix.."""
    if not line.startswith(tuple(todo_prefixes())):
        return Err(
            "Magic todos must satisfy one of; (1) Have a non-default"
            " (i.e. not 'O') priority set (2) Have been marked complete"
            " with an 'x' prefix (3) Have been marked open with an 'o'"
            f" prefix. line={line!r}"
        )

    return Ok(None)


@todo_spell
def x_tag(todo: T) -> T:
    """Handles tags of the form x:1234 where 1234 is the current time."""
    if todo.metadata is None or "x" not in todo.metadata:
        return todo

    dtime = todo.metadata["x"]
    desc = " ".join(todo.desc.split(" ")[1:]) + f" dtime:{dtime}"

    new_todo = todo.new(desc=desc, done=True)
    line = new_todo.to_line()
    return type(todo).from_line(line).unwrap()


@todo_spell
def group_tags(todo: T) -> T:
    """Groups all @ctxs, +projs, and meta:data at the end of the line."""
    if not (todo.contexts or todo.projects or todo.metadata):
        return todo

    all_words = [w for w in todo.desc.split(" ") if w != "|"]
    new_words = []
    regular_words_found = False
    for i, word in enumerate(all_words[:]):
        all_prev_words_are_tags = _all_words_are_tags(all_words[:i])
        all_next_words_are_tags = _all_words_are_tags(all_words[i + 1 :])
        is_edge_tag = all_next_words_are_tags or all_prev_words_are_tags

        if is_metadata_tag(word) and word[-1] in PUNCTUATION:
            return todo

        if is_any_prefix_tag(word) and (
            word[-1] in PUNCTUATION or not all_next_words_are_tags
        ):
            if regular_words_found:
                new_words.append(word[1:])
            continue

        if is_any_prefix_tag(word) and is_edge_tag:
            continue

        if is_metadata_tag(word):
            continue

        regular_words_found = True
        new_words.append(word)

    if not regular_words_found:
        return todo

    desc = " ".join(new_words).strip()
    desc += " |"

    if todo.contexts:
        desc += " " + " ".join(
            CONTEXT_PREFIX + ctx for ctx in sorted(todo.contexts)
        )

    if todo.projects:
        desc += " " + " ".join(
            PROJECT_PREFIX + ctx for ctx in sorted(todo.projects)
        )

    if todo.metadata:
        desc += " " + " ".join(
            f"{k}:{v}" for (k, v) in sorted(todo.metadata.items())
        )

    return todo.new(desc=desc)


def _all_words_are_tags(words: Iterable[str]) -> bool:
    """Returns True if all `words` are special words."""
    return all(is_any_tag(w) for w in words)


@todo_spell
def add_create_date(todo: T) -> T:
    """Adds today's date as the create date for this Todo."""
    if todo.create_date is not None:
        return todo

    today = dt.date.today()
    return todo.new(create_date=today)


@todo_spell
def add_done_date(todo: T) -> T:
    """Adds today's date as the done date for this Todo (if done)."""
    if todo.done_date is not None:
        return todo

    if not todo.done:
        return todo

    today = dt.date.today()
    return todo.new(done_date=today)


@todo_spell
def add_ctime(todo: T) -> T:
    """Adds creation time to T via the 'ctime' metadata tag."""
    if "ctime" in todo.metadata:
        return todo

    now = dt.datetime.now()

    metadata = dict(todo.metadata)
    metadata["ctime"] = f"{now.hour:0>2}{now.minute:0>2}"

    return todo.new(metadata=metadata)


@to_line_spell
def add_o_prefix(line: str) -> str:
    """Adds the 'o ' prefix to the Todo line."""
    if line.startswith(todo_prefixes()):
        return line

    return O_PREFIX + line


@from_line_spell
def remove_o_prefix(line: str) -> str:
    """Removes the 'o ' prefix from the Todo line."""
    if not line.startswith(O_PREFIX):
        return line

    return line[len(O_PREFIX) :]


@to_line_spell
def add_x_prefix(line: str) -> str:
    """Adds the 'x:HHMM ' prefix to the Todo line (when done)."""
    if not line.startswith("x "):
        return line

    words = line.split(" ")[1:]
    for i, word in enumerate(words[:]):
        if word.startswith("dtime:"):
            del words[i]
            dtime = word.split(":")[1]
            break
    else:
        now = dt.datetime.now()
        dtime = f"{now.hour:0>2}{now.minute:0>2}"

    rest = " ".join(words)
    return f"x:{dtime} {rest}"


@from_line_spell
def remove_x_prefix(line: str) -> str:
    """Removes the 'x:HHMM ' prefix from the Todo line."""
    if not line.startswith("x:"):
        return line

    xhhmm, *words = line.split(" ")
    dtime = xhhmm.split(":")[1]

    rest = " ".join(words)
    return f"x {rest} dtime:{dtime}"



"""Contains the TodoGroup class definition."""

from __future__ import annotations

import datetime as dt
from logging import Logger
from pathlib import Path
from typing import Generic, Iterable, Iterator, Mapping, Type

from eris import Err
from metaman import cname
from typist import PathLike

from .types import MetadataChecker, Priority, T


logger = Logger(__name__)


class TodoGroup(Generic[T]):
    """Manages a group of Todo objects."""

    def __init__(
        self, todo_map: dict[str, T], path_map: dict[str, Path]
    ) -> None:
        self.todo_map = todo_map
        self.path_map = path_map

        self.todos = list(todo_map.values())

    def __repr__(self) -> str:  # noqa: D105
        return f"{cname(self)}({self.todos})"

    def __iter__(self) -> Iterator[T]:
        """Yields the Todo objects that belong to this group."""
        yield from self.todos

    def __len__(self) -> int:  # noqa: D105
        return len(self.todos)

    @classmethod
    def from_path(
        cls,
        todo_type: Type[T],
        path_like: PathLike,
        *,
        path_map: dict[str, Path] = None,
        todo_map: dict[str, T] = None,
    ) -> TodoGroup:
        """Reads all todo lines from a given file or directory (recursively).

        Pre-conditions:
            * `path_like` exists and is either a file or directory.
        """
        path = Path(path_like)
        if path_map is None:
            path_map = {}

        if todo_map is None:
            todo_map = {}

        assert path.exists(), f"The provided path does not exist: {path}"

        if path.is_file():
            logger.debug(
                "Attempting to load todos from text file: file=%r", str(path)
            )

            todos = []
            for line in path.read_text().split("\n"):
                line = line.strip()
                todo_result = todo_type.from_line(line)
                if isinstance(todo_result, Err):
                    continue

                todo = todo_result.ok()
                logger.debug("New todo loaded: todo=%r", todo)
                todos.append(todo)

                key = todo.ident
                path_map[key] = path
                todo_map[key] = todo

        else:
            assert path.is_dir(), (
                "The provided path exists but is neither a file nor a"
                f" directory: {path}"
            )

            logger.debug("Loading from directory: dir=%r", str(path))
            for other_path in path.glob("*"):
                if other_path.is_file() and other_path.suffix != ".txt":
                    continue

                TodoGroup.from_path(
                    todo_type, other_path, path_map=path_map, todo_map=todo_map
                )

        return cls(todo_map, path_map)

    def to_disk(self, file_path: PathLike) -> None:
        """Writes a group of Todos into `file_path`."""
        file_path = Path(file_path)
        assert file_path.is_file()
        file_path.write_text(
            "\n".join(todo.to_line() for todo in sorted(self))
        )

    def filter_by(
        self,
        *,
        contexts: Iterable[str] = (),
        create_date: dt.date = None,
        desc: str = None,
        done_date: dt.date = None,
        done: bool = None,
        metadata_checks: Mapping[str, MetadataChecker] = None,
        priorities: Iterable[Priority] = (),
        projects: Iterable[str] = (),
    ) -> TodoGroup:
        """Filter this group using one or more Todo properties."""
        path_map = {}
        todo_map = {}
        for key, todo in self.todo_map.items():
            skip_this_todo = False
            for ctx in contexts:
                if ctx.startswith("-") and ctx[1:] in todo.contexts:
                    skip_this_todo = True
                    break

                if ctx not in todo.contexts:
                    skip_this_todo = True
                    break

            for proj in projects:
                if proj.startswith("-") and proj[1:] in todo.projects:
                    skip_this_todo = True
                    break

                if proj not in todo.projects:
                    skip_this_todo = True
                    break

            if priorities and todo.priority not in priorities:
                continue

            if create_date is not None and todo.create_date != create_date:
                continue

            if desc is not None and desc.lower() not in todo.desc.lower():
                continue

            if done_date is not None and todo.done_date != done_date:
                continue

            if done is not None and todo.done != done:
                continue

            if metadata_checks is not None:
                for mkey, check in metadata_checks.items():
                    if mkey not in todo.metadata:
                        skip_this_todo = True
                        break

                    mvalue = todo.metadata[mkey]
                    assert isinstance(mvalue, str)
                    if not check(mvalue):
                        skip_this_todo = True
                        break

            if skip_this_todo:
                continue

            path_map[key] = self.path_map[key]
            todo_map[key] = todo

        return TodoGroup(todo_map, path_map)
